# -*- coding: utf-8 -*-
import os
import warnings
import logging
from functools import partial

from six.moves.urllib.parse import unquote, urlparse
from six import string_types

import rqdatac


if getattr(os, "register_at_fork", None):
    def _close_after_fork():
        global _CLIENT
        if _CLIENT is not _DUMMY and _CLIENT.PID != os.getpid():
            reset()

    os.register_at_fork(after_in_child=_close_after_fork)


class DummyClient:
    PID = -1

    def execute(self, *args, **kwargs):
        raise RuntimeError("rqdatac is not initialized")

    def reset(self):
        pass

    def info(self):
        print('rqdatac is not initialized')

    def close(self):
        pass

    execute_with_timeout = execute


_DUMMY = DummyClient()

_CLIENT = _DUMMY


def reset():
    """reset all collection to a server.
    this function is helpful when you fork a inited client
    """
    _CLIENT.reset()
    _CLIENT.PID = os.getpid()


def initialized():
    """
    is rqdatac already initialized ?
    :return: Bool
    """
    global _CLIENT, _DUMMY
    return _CLIENT is not _DUMMY


def _set_sock_factory(proxy_info):
    from rqdatac.connection import Connection
    if proxy_info is None:
        import socket
        Connection.set_sock_factory(socket.create_connection)
        return

    try:
        import socks
    except ImportError:
        raise RuntimeError("PySocks is required when use proxy. You can install it using `pip install PySocks` or "
                           "`pip install rqdatac[proxy]`")
    if not isinstance(proxy_info, tuple) or len(proxy_info) != 5:
        raise ValueError("expected a tuple like (proxy_type, host, port, user, password)")
    proxy_type, host, port, user, password = proxy_info
    if proxy_type.upper() == "HTTP":
        proxy_type = socks.PROXY_TYPE_HTTP
    elif proxy_type.upper() == "SOCKS4":
        proxy_type = socks.PROXY_TYPE_SOCKS4
    elif proxy_type.upper() == "SOCKS5":
        proxy_type = socks.PROXY_TYPE_SOCKS5
    else:
        raise ValueError("proxy type {} not supported".format(proxy_type))

    Connection.set_sock_factory(partial(socks.create_connection, proxy_type=proxy_type, addr=host, port=port,
                                        username=user, password=password))


def init(username=None, password=None, addr=("rqdatad-pro.ricequant.com", 16011), *_, **kwargs):
    """init to a remote server

    rqdatac connection is thread safe but not fork safe. Every thread have their own connection by
    default. you can set param 'use_pool' to True to use a connection pool instead.

    NOTE: if you are using rqdatac with python < 3.7 in a multi-process program, remember to call
    reset in child process.

    Environment Variable: RQDATAC_CONF / RQDATAC2_CONF. When init called with no argument, the value in
    RQDATAC2_CONF is used as uri, then RQDATAC_CONF.

    :param username: string
    :param password: string
    :param addr: ('127.0.0.1', 80) or '127.0.0.1:80'

    :param uri: keyword only. a uri like rqdata://username:password@host:port or tcp://username:password@host:port
    :param connect_timeout: socket connect connect timeout, default is 5 sec.
    :param timeout: socket time out, default is 60 sec.
    :param lazy: True by default, means "do not connect to server immediately".
    :param use_pool: use connection pool. default is False
    :param max_pool_size: max pool size, default is 8
    :param proxy_info: a tuple like (proxy_type, host, port, user, password) if use proxy, default is None

    """
    extra_args = {k: kwargs.pop(k) for k in ('timeout', 'connect_timeout') if k in kwargs}
    _set_sock_factory(kwargs.pop("proxy_info", None))
    logging.getLogger("rqdata").disabled = not kwargs.pop('debug', False)

    uri = kwargs.pop("uri", None)
    if not (username or password or uri):
        uri = os.environ.get("RQDATAC2_CONF") or os.environ.get("RQDATAC_CONF")
    if username and password and addr:
        addr = parse_address(addr)
    elif uri:
        r = urlparse(unquote(uri))
        username = username or r.username
        password = password or r.password
        addr = parse_address((r.hostname, r.port))
    else:
        raise ValueError("username/password/addr or uri expected")

    global _CLIENT
    if _CLIENT is not _DUMMY:
        warnings.warn("rqdatac is already inited. Settings will be changed.", stacklevel=0)
        reset()

    auth_info = {'username': username, 'password': password, 'ver': rqdatac.__version__}
    if kwargs.pop("use_pool", False):
        from .connection_pool import ConnectionPool
        max_pool_size = kwargs.pop("max_pool_size", 8)
        _CLIENT = ConnectionPool(addr, auth=auth_info, max_pool_size=max_pool_size, **extra_args)
    else:
        from .thread_local import ThreadLocalConnection
        _CLIENT = ThreadLocalConnection(addr, auth=auth_info, **extra_args)

    _CLIENT.PID = os.getpid()

    def show_info():
        print('rqdatac version:', rqdatac.__version__)
        print('server address: {}:{}'.format(addr[0], addr[1]))
        if username == 'license':
            print('You are using license:\r\n{}'.format(password))
        elif username.startswith('rqpro.'):
            print('You are using a RQPro account: {}'.format(username.split('.', 1)[1]))
        elif username == 'sid':
            print('You are using your RQPro account')
        else:
            print('You are using account: {}'.format(username))
    _CLIENT.info = show_info

    if username == "license":
        quota = get_client().execute("user.get_quota")
        remaining_days = quota["remaining_days"]
        is_trial = quota["license_type"] == "TRIAL"
        if is_trial or 0 <= remaining_days <= 14:
            warnings.warn("Your account will be expired after  {} days. "
                          "Please call us at 0755-22676337 to upgrade or purchase or "
                          "renew your contract.".format(remaining_days))
    elif not kwargs.get("lazy", True):
        get_client().execute("get_all_trading_dates")


def get_client():
    if _CLIENT.PID != os.getpid():
        reset()
    return _CLIENT


def parse_address(address):
    if isinstance(address, tuple):
        host, port = address
        return host, int(port)
    elif isinstance(address, string_types):
        if ":" not in address:
            return address
        host, port = address.rsplit(":", 1)
        return host, int(port)
    else:
        raise ValueError("address must be a str or tuple")
